#import <OpenGL/gl.h>
#import <OpenGL/OpenGL.h>

#import "GLSplinePatch.h"
#import "GLSplinePatchUI.h"

@implementation GLSplinePatch : QCPatch

+ (int)executionMode
{
	return 1;
}
+ (BOOL)allowsSubpatches
{
	return NO;
}
+ (int)timeMode
{
	return 1;
}

+ (Class)inspectorClassWithIdentifier:(id)fp8
{
	return [GLSplinePatchUI class];
}

- (id)initWithIdentifier:(id)fp8
{
	id z=[super initWithIdentifier:fp8];
	if(z)
	{
		// start with a sane tension
		[inputTension setDoubleValue:2.0];
		// start with a nice step
		[inputSubdivisions setIndexValue:200];

		[inputShowPoints setBooleanValue:TRUE];

		[inputWidth setMinDoubleValue:0.4];
		[inputWidth setDoubleValue:1.0];

		// start out with 3 points, so we show a nice spline
		[self addPoint];
		[self addPoint];
		[self addPoint];
		[xArray[0] setDoubleValue:-0.5];
		[yArray[1] setDoubleValue: 0.5];
		[xArray[2] setDoubleValue: 0.5];
		[colorArray[0] setRed:0.7 green:0.4 blue:0.4 alpha:0.9];
		[colorArray[1] setRed:0.4 green:0.4 blue:0.7 alpha:0.9];
		[colorArray[2] setRed:0.4 green:0.7 blue:0.4 alpha:0.9];
	}
	return z;
}
- (id)attributes
{
	NSMutableDictionary *at=[[super attributes] mutableCopy];
	[at setObject:@"Kineme GL Spline" forKey:@"name"];
	return at;
}


- (id)setup:(QCOpenGLContext *)context
{
	return context;
}
- (void)cleanup:(QCOpenGLContext *)context
{
}


- (void)enable:(QCOpenGLContext *)context
{
}
- (void)disable:(QCOpenGLContext *)context
{
}


- (BOOL)execute:(QCOpenGLContext *)context time:(double)time arguments:(NSDictionary *)arguments
{
	if(controlPoints < 1)
		return YES;

	CGLSetCurrentContext([context CGLContextObj]);
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
	glLineWidth([inputWidth doubleValue]);
	
	unsigned int i;

	for(i=0;i<controlPoints;++i)
	{
		ptsx[i] = [xArray[i] doubleValue];
		ptsy[i] = [yArray[i] doubleValue];
		ptsz[i] = [zArray[i] doubleValue];		
	}
	
	double t;
	// this isn't technically necessary, but good form
	if([inputSubdivisions indexValue] == 0)
		[inputSubdivisions setIndexValue:1];
		
	double steps = 1.0/(double)[inputSubdivisions indexValue];	// move some message stuff out of the loop
	
	double px, py, pz;
	double B0, B1, B2, B3;
	
	px = ptsx[0];
	py = ptsy[0];
	pz = ptsz[0];
	
	glBegin(GL_LINES);
	
	float a = (float)[inputTension doubleValue];
		
	int i0, i1, i2;
	float r1, g1, b1, a1, r2, g2, b2, a2;
	
	[colorArray[0] getRed:&r2 green:&g2 blue:&b2 alpha:&a2];
	
	for (i = 0; i < controlPoints - 1; i++)
	{
		glVertex3f(ptsx[i], ptsy[i], ptsz[i]);

		i0 = (i == 0) ? controlPoints - 1 : i - 1;
		i1 = (i == controlPoints - 1) ? 0 : i + 1;
		i2 = (i >= controlPoints - 2) ? ((i == controlPoints - 1) ? 1 : 0) : i + 2;
		
		// ...move more messages out of the inner loop...
		r1=r2;
		g1=g2;
		b1=b2;
		a1=a2;
		[colorArray[i+1] getRed:&r2 green:&g2 blue:&b2 alpha:&a2];
		
		for (t = 0; t <= 1; t+= steps)
		{
			/* several years ago(1999-2000ish?), I found that CSE'ing spline equations like these yielded signifigant performance benefits.
				This probably isn't the case anymore, but it might be a fun experiment.
				
				for example, t*t is used twice, as is (1-t)*(1-t).  I'm hoping modern CSE optimization passes catch this already...*/
			B0 = (1-t)*(1-t)*(1-t);
			B1 = 3*t*(1-t)*(1-t);
			B2 = 3*t*t*(1-t);
			B3 = t*t*t;
					
			px = ptsx[i]*B0 + (ptsx[i] + (ptsx[i1] - ptsx[i0])/a)*B1 + (ptsx[i1] - (ptsx[i2] - ptsx[i])/a)*B2 + ptsx[i1]*B3;
			py = ptsy[i]*B0 + (ptsy[i] + (ptsy[i1] - ptsy[i0])/a)*B1 + (ptsy[i1] - (ptsy[i2] - ptsy[i])/a)*B2 + ptsy[i1]*B3;
			pz = ptsz[i]*B0 + (ptsz[i] + (ptsz[i1] - ptsz[i0])/a)*B1 + (ptsz[i1] - (ptsz[i2] - ptsz[i])/a)*B2 + ptsz[i1]*B3;
			glColor4f(
				r1*(1.0-t) + r2*t,
				g1*(1.0-t) + g2*t,
				b1*(1.0-t) + b2*t,
				a1*(1.0-t) + a2*t);
			glVertex3f(px, py, pz);
			glVertex3f(px, py, pz);
		}

		glVertex3f(ptsx[i1], ptsy[i1], ptsz[i1]);
	}
	
	glEnd();

	if([inputShowPoints booleanValue] == TRUE)
	{
		glPointSize(8);
		glEnable(GL_POINT_SMOOTH);
		glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);
		glBegin(GL_POINTS);
			for(i=0;i<controlPoints;++i)
			{
				glColor4f([colorArray[i] redComponent], [colorArray[i] greenComponent], [colorArray[i] blueComponent],[colorArray[i] alphaComponent]);
				glVertex3f(ptsx[i],ptsy[i],ptsz[i]);
			}
		glEnd();
	}

	return YES;
}

- (void)addPoint
{
	if(controlPoints<40)	// if we have more than 40 points, the patch doesn't fit within the composition editor..
	{
		ptsx = (double*)realloc(ptsx,(sizeof(double)*(controlPoints+1)));
		ptsy = (double*)realloc(ptsy,(sizeof(double)*(controlPoints+1)));
		ptsz = (double*)realloc(ptsz,(sizeof(double)*(controlPoints+1)));
		
		xArray = (QCNumberPort**)realloc(xArray, ((sizeof(QCNumberPort*)) * (controlPoints+1)));
		yArray = (QCNumberPort**)realloc(yArray, ((sizeof(QCNumberPort*)) * (controlPoints+1)));
		zArray = (QCNumberPort**)realloc(zArray, ((sizeof(QCNumberPort*)) * (controlPoints+1)));
		colorArray = (QCColorPort**)realloc(colorArray, ((sizeof(QCColorPort*)) * (controlPoints+1)));
		
		xArray[controlPoints] = [self createInputWithPortClass:[QCNumberPort class] forKey:[NSString stringWithFormat:@"X%i Position",controlPoints+1] attributes:nil];
		yArray[controlPoints] = [self createInputWithPortClass:[QCNumberPort class] forKey:[NSString stringWithFormat:@"Y%i Position",controlPoints+1] attributes:nil];
		zArray[controlPoints] = [self createInputWithPortClass:[QCNumberPort class] forKey:[NSString stringWithFormat:@"Z%i Position",controlPoints+1] attributes:nil];
		colorArray[controlPoints] = [self createInputWithPortClass:[QCColorPort class] forKey:[NSString stringWithFormat:@"Color %i",controlPoints+1] attributes:nil];
		
		[xArray[controlPoints] retain];
		[yArray[controlPoints] retain];
		[zArray[controlPoints] retain];
		[colorArray[controlPoints] retain];
		
		++controlPoints;
	}
}

- (void)removePoint
{
	if(controlPoints>2)
	{		
		xArray = (QCNumberPort**)realloc(xArray, ((sizeof(QCNumberPort*)) * (controlPoints-1)));
		yArray = (QCNumberPort**)realloc(yArray, ((sizeof(QCNumberPort*)) * (controlPoints-1)));
		zArray = (QCNumberPort**)realloc(zArray, ((sizeof(QCNumberPort*)) * (controlPoints-1)));
		colorArray = (QCColorPort**)realloc(colorArray, ((sizeof(QCColorPort*)) * (controlPoints-1)));
		
		ptsx = (double*)realloc(ptsx,(sizeof(double)*(controlPoints-1)));
		ptsy = (double*)realloc(ptsy,(sizeof(double)*(controlPoints-1)));
		ptsz = (double*)realloc(ptsz,(sizeof(double)*(controlPoints-1)));

		[self deleteInputForKey:[NSString stringWithFormat:@"X%i Position",controlPoints]];
		[self deleteInputForKey:[NSString stringWithFormat:@"Y%i Position",controlPoints]];
		[self deleteInputForKey:[NSString stringWithFormat:@"Z%i Position",controlPoints]];
		[self deleteInputForKey:[NSString stringWithFormat:@"Color %i",controlPoints]];

		[xArray[controlPoints-1] release];
		[yArray[controlPoints-1] release];
		[zArray[controlPoints-1] release];
		[colorArray[controlPoints-1] release];
		
		--controlPoints;
	}
}

- (NSDictionary*)state
{
	int i;
	float r,g,b,a;

	NSMutableDictionary *stateDict = [[NSMutableDictionary alloc] initWithCapacity:1+controlPoints*3];
	[stateDict addEntriesFromDictionary:[super state]];
	
	[stateDict setObject:[NSNumber numberWithInt:controlPoints] forKey:@"controlPoints"];
	
	for(i=0;i<controlPoints;++i)
	{
		[colorArray[i] getRed:&r green:&g blue:&b alpha:&a];
		[stateDict setObject:[NSNumber numberWithDouble: ptsx[i]] forKey:[NSString stringWithFormat:@"x%i",i]];
		[stateDict setObject:[NSNumber numberWithDouble: ptsy[i]] forKey:[NSString stringWithFormat:@"y%i",i]];
		[stateDict setObject:[NSNumber numberWithDouble: ptsz[i]] forKey:[NSString stringWithFormat:@"z%i",i]];
		[stateDict setObject:[NSNumber numberWithDouble: r ]      forKey:[NSString stringWithFormat:@"r%i",i]];
		[stateDict setObject:[NSNumber numberWithDouble: g ]      forKey:[NSString stringWithFormat:@"g%i",i]];
		[stateDict setObject:[NSNumber numberWithDouble: b ]      forKey:[NSString stringWithFormat:@"b%i",i]];
		[stateDict setObject:[NSNumber numberWithDouble: a ]      forKey:[NSString stringWithFormat:@"a%i",i]];
	}

	[stateDict autorelease];
	return stateDict;
}

- (BOOL)setState:(NSDictionary*)state
{
	int i;

	/* clean up existing state, if it exists (init creates some) */
	for(i=0;i<controlPoints;++i)
	{
		[xArray[i] release];
		[yArray[i] release];
		[zArray[i] release];
		[colorArray[i] release];
		
		[self deleteInputForKey:[NSString stringWithFormat:@"X%i Position",i+1]];
		[self deleteInputForKey:[NSString stringWithFormat:@"Y%i Position",i+1]];
		[self deleteInputForKey:[NSString stringWithFormat:@"Z%i Position",i+1]];
		[self deleteInputForKey:[NSString stringWithFormat:@"Color %i",i+1]];
	}
	
	controlPoints = [[state objectForKey:@"controlPoints"] intValue];
	
	/*if(controlPoints < 2)
		return NO;*/
	
	ptsx = (double*)realloc(ptsx,(sizeof(double)*(controlPoints)));
	ptsy = (double*)realloc(ptsy,(sizeof(double)*(controlPoints)));
	ptsz = (double*)realloc(ptsz,(sizeof(double)*(controlPoints)));
	
	xArray = (QCNumberPort**)realloc(xArray, ((sizeof(QCNumberPort*)) * (controlPoints)));
	yArray = (QCNumberPort**)realloc(yArray, ((sizeof(QCNumberPort*)) * (controlPoints)));
	zArray = (QCNumberPort**)realloc(zArray, ((sizeof(QCNumberPort*)) * (controlPoints)));
	colorArray = (QCColorPort**)realloc(colorArray, ((sizeof(QCColorPort*)) * (controlPoints+1)));

	for(i=0;i<controlPoints;++i)
	{
		xArray[i] = [self createInputWithPortClass:[QCNumberPort class] forKey:[NSString stringWithFormat:@"X%i Position",i+1] attributes:nil];
		yArray[i] = [self createInputWithPortClass:[QCNumberPort class] forKey:[NSString stringWithFormat:@"Y%i Position",i+1] attributes:nil];
		zArray[i] = [self createInputWithPortClass:[QCNumberPort class] forKey:[NSString stringWithFormat:@"Z%i Position",i+1] attributes:nil];
		colorArray[i] = [self createInputWithPortClass:[QCColorPort class] forKey:[NSString stringWithFormat:@"Color %i",i+1] attributes:nil];
		
		[xArray[i] setDoubleValue:[[state objectForKey:[NSString stringWithFormat:@"x%i",i]] doubleValue]];
		[yArray[i] setDoubleValue:[[state objectForKey:[NSString stringWithFormat:@"y%i",i]] doubleValue]];
		[zArray[i] setDoubleValue:[[state objectForKey:[NSString stringWithFormat:@"z%i",i]] doubleValue]];
		[colorArray[i] setRed:	[[state objectForKey:[NSString stringWithFormat:@"r%i",i]] doubleValue]
						green:	[[state objectForKey:[NSString stringWithFormat:@"g%i",i]] doubleValue]
						blue:	[[state objectForKey:[NSString stringWithFormat:@"b%i",i]] doubleValue]
						alpha:	[[state objectForKey:[NSString stringWithFormat:@"a%i",i]] doubleValue]];
						
		[xArray[i] retain];
		[yArray[i] retain];
		[zArray[i] retain];
		[colorArray[i] retain];
	}
	// this restore default ports and position.
	[super setState:state];
	
	return YES;
}

@end
